@@ -1,6 +1,6 @@
 import base64
 import re
-from typing import List, Dict, Union, Callable
+from typing import List, Dict, Set, Union, Callable
 
 class HELML:
 
@@ -17,10 +17,15 @@ class HELML:
     CUSTOM_FORMAT_DECODER = None
     CUSTOM_VALUE_ENCODER = None
     CUSTOM_VALUE_DECODER = None
+    
+    ENABLE_SPC_IDENT = True # For encode: add space-indentation at begin of string
+    ENABLE_BONES = True # For encode: enable use "next"-keys like :--:
+    ENABLE_KEY_UPLINES = True # For encode: add empty line before create-array-keys
+    ENABLE_HASHSYMBOLS = True # For encode: adding # after nested-blocks
 
     @staticmethod
     def encode(
-        arr: Union[list, dict, tuple],
+        arr: Union[dict, list, tuple, set],
         url_mode: bool = False
     ) -> str:
         """
@@ -28,39 +33,32 @@ class HELML:
         
         :param arr: Input data structure (list, dict, or tuple) to be encoded.
         :param url_mode: A boolean indicating if the URL mode should be used.
-        :param val_encoder: A function to encode values or True for default encoding.
         :return: Encoded HELML string.
         """
         results_arr = []
-        if not isinstance(arr, (list, dict, tuple)):
-            raise ValueError("List or dictionary required")
 
         str_imp = "~" if url_mode else "\n"
         lvl_ch = "." if url_mode else ":"
         spc_ch = "_" if url_mode else " "
 
-        HELML._encode(arr, results_arr, 0, lvl_ch, spc_ch)
+        is_list = HELML.num_keys_cnt(arr)
+        
+        HELML._encode(arr, results_arr, 0, lvl_ch, spc_ch, is_list)
+
+        if url_mode and len(results_arr) == 1:
+            results_arr.append('')
 
         return str_imp.join(results_arr)
     
     @staticmethod
     def _encode(
-        arr: Union[dict, list, tuple],
+        arr: Union[dict, list, tuple, set],
         results_arr: list,
         level: int = 0,
         lvl_ch: str = ":",
-        spc_ch: str = " "
+        spc_ch: str = " ",
+        is_list: int = 0
     ) -> None:
-        """
-        Encode arr to results_arr (by reference).
-        
-        :param arr: Input data structure (dict, list, or tuple) to be encoded.
-        :param results_arr: A list of strings representing the encoded data.
-        :param val_encoder: A function to encode values or True for default encoding.
-        :param level: The current level of nesting.
-        :param lvl_ch: Character used for level separation.
-        :param spc_ch: Character used for space.
-        """
 
         # select value-encoder function: custom or internal default
         value_enco_fun = HELML.CUSTOM_VALUE_ENCODER if HELML.CUSTOM_VALUE_ENCODER is not None else HELML.valueEncoder
@@ -73,78 +71,98 @@ class HELML:
             if not isinstance(key, str):
                 key = str(key)
 
-            # get first char
-            first_char = key[0]
-            # encode key in base64url if it contains unwanted characters
-            if lvl_ch in key or "~" in key or first_char == "#" or first_char == spc_ch or first_char == ' ':
-                first_char = "-"
-            if first_char == "-" or key[-1] == spc_ch or key[-1] == ' ' or not all(c.isprintable() for c in key):
-                # add "-" to the beginning of the key to indicate it's in base64url
-                key = "-" + HELML.base64url_encode(key)
+            if is_list and HELML.ENABLE_BONES and spc_ch == ' ':
+                key = '--'
+            elif not is_list:
+                # encode key to base64url if contains unwanted characters
+                # check first char
+                fc = key[:1]
+                # encode key in base64url if it contains unwanted characters
+                if lvl_ch in key or fc == "#" or fc == spc_ch or fc == ' ' or fc == ''  or key[-1] == spc_ch or key[-1] == ' ':
+                    fc = '-'
+                elif not (re.match(r'^[ -}]+$', key) if spc_ch == '_' else re.match(r'^[^\x00-\x1F\x7E-\xFF]+$', key)):
+                    fc = '-'
+                if fc == '-':
+                    # add "-" to the beginning of the key to indicate it's in base64url
+                    key = '-' + HELML.base64Uencode(key)
 
             # add the appropriate number of colons to the left of the key, based on the current level
-            key = lvl_ch * level + key
+            ident = lvl_ch * level
 
-            if isinstance(value, (list, dict, tuple)):
-                # if the value is a dictionary, call this function recursively and increase the level
-                if isinstance(value, (list, tuple)):
+             # add space-ident to the left of the key (if need)
+            if HELML.ENABLE_SPC_IDENT and spc_ch == ' ':
+                ident = spc_ch * level + ident
+
+            is_arr = isinstance(value, (list, tuple, set))
+
+            if is_arr or isinstance(value, dict):
+                # Add empty line before create-key
+                if HELML.ENABLE_KEY_UPLINES and spc_ch == ' ':
+                    results_arr.append('')
+                
+                # add ":" after key for lists
+                if is_arr:
                     key += lvl_ch
-                results_arr.append(key)
-                HELML._encode(value, results_arr, level + 1, lvl_ch, spc_ch)
+
+                results_arr.append(ident + key)
+
+                HELML._encode(value, results_arr, level + 1, lvl_ch, spc_ch, is_arr)
+
+                if HELML.ENABLE_KEY_UPLINES and spc_ch == ' ':
+                    results_arr.append(spc_ch * level + '#')
             else:
                 # use selected value encoder function
                 value = value_enco_fun(value, spc_ch)
 
                 # add the key:value pair to the output
-                results_arr.append(key + lvl_ch + value)
+                results_arr.append(ident + key + lvl_ch + value)
 
     @staticmethod
     def decode(
-        src_rows: Union[str, List[str], Dict[str, str]]
+        src_rows: str,
+        get_layers: Union[int, str, Set[Union[str, int]], List[Union[str, int]]] = [0]
     ) -> Dict:
-        """
-        Decodes a HELML formatted string or list of strings into a nested dictionary.
-
-        Args:
-            src_rows: The HELML input as a string or strings-array.
-            val_decoder: The custom value decoder function. For default encoder set True.
-
-        Returns:
-            Dict: The decoded nested dictionary.
-        """
 
         # select value decoder function custom or internal default
         value_deco_fun = HELML.CUSTOM_VALUE_DECODER if HELML.CUSTOM_VALUE_DECODER is not None else HELML.valueDecoder
 
-        lvl_ch = ":"
-        spc_ch = " "
-        # If the input is an array, use it. Otherwise, split the input string into an array.
-        if isinstance(src_rows, (list, dict)):
-            if isinstance(src_rows, dict):
-                str_arr = list(src_rows)
-            else:
-                str_arr = src_rows
+        lvl_ch: str = ":"
+        spc_ch: str = " "
+        layer_init: str = '0'
+        layer_curr:str = layer_init
+        all_layers = set([layer_init])
 
-        elif isinstance(src_rows, str):
-            for exploder_ch in ["\n", "\r", "~"]:
-                if exploder_ch in src_rows:
-                    if "~" == exploder_ch:
-                        lvl_ch = "."
-                        spc_ch = "_"
-                    break
+        # Prepare layers_set from get_layers
+        # 1. Modify get_layers if needed: convert single T to array [0, T]
+        if isinstance(get_layers, (int, str)):
+            get_layers = [layer_init, get_layers]
+        # 2. Create layers_set and set all values from get_layers (str)
+        layers_set = set()
+        for i in get_layers:
+            if not isinstance(i, str):
+                i = str(i)
+            layers_set.add(i)
 
-            str_arr = src_rows.split(exploder_ch)
 
-        else:
-            raise ValueError("Array or String required")
+        # Explode src_rows to lines
+        for exploder_ch in ["\n", "\r", "~"]:
+            if exploder_ch in src_rows:
+                if "~" == exploder_ch:
+                    lvl_ch = "."
+                    spc_ch = "_"
+                break
+
+        str_arr = src_rows.split(exploder_ch)
 
         # Initialize result array and stack for keeping track of current array nesting
         result = {}
         stack = []
 
-        # array of stack stamps for delayed conversion of dict to list
+        # array of stack stamps for delayed conversion dict to list
         tolist = []
 
+        min_level: int = -1
+
         # Loop through each line in the input array
         for line in str_arr:
             line = line.strip()
@@ -168,28 +186,57 @@ class HELML:
             key = parts[0] if parts[0] else 0
             value = parts[1] if len(parts) > 1 else None
 
-            # Decode the key if it starts with an equals sign
-            if isinstance(key, str) and key.startswith("-"):
-                key = HELML.base64url_decode(key[1:])
-                if not key:
-                    key = "ERR"
+            # check min_level
+            if min_level < 0 or min_level > level:
+                min_level = level
 
-            # Remove keys from the stack until it matches the current level
-            while len(stack) > level:
-                stack.pop()
+            extra_keys_cnt: int = len(stack) - (level - min_level)
+            if extra_keys_cnt > 0:
+                layer_curr = layer_init
+                while len(stack) and extra_keys_cnt > 0:
+                    stack.pop()
+                    extra_keys_cnt -= 1
 
             # Find the parent element in the result dictionary for the current key
             parent = result
             for parent_key in stack:
                 parent = parent[parent_key]
 
+            # Decode the key if it starts with an equals sign
+            if key.startswith("-"):
+                # Next number keys
+                if key == '--' or key == '---':
+                    # Next Num keys
+                    key = (str)(len(parent))
+
+                # Layer control keys
+                elif key == '-+' or key == '-++':
+                    if not value is None:
+                        value = value.strip()
+                    if key == '-++':
+                        layer_init = value if value else '0'
+                        layer_curr = layer_init
+                    elif key == '-+':
+                        if value is None:
+                            layer_curr = str(int(layer_curr) + 1) if layer_curr.isdigit() else layer_init
+                        else:
+                            layer_curr = layer_init if value == '' else value
+
+                    all_layers.add(layer_curr)
+                    continue
+                else:
+                    decoded_key = HELML.base64Udecode(key[1:])
+                    if decoded_key != None:
+                        key = decoded_key
+
             # If the value is null, start a new dictionary and add it to the parent dictionary
             if value is None or value == '':
+                layer_curr = layer_init
                 parent[key] = {}
                 stack.append(key)
                 if value == '':
                     tolist.append(stack.copy())
-            else:
+            elif layer_curr in layers_set:
                 # Decode the value by selected value-decoder-function
                 value = value_deco_fun(value, spc_ch)
                 # Add the key-value pair to the current dictionary
@@ -205,10 +252,22 @@ class HELML:
             if isinstance(parent, list) and HELML.is_numeric(last_key) and int(last_key) < len(parent):
                 last_key = int(last_key)
             if ((isinstance(parent, dict) and parent.get(last_key) is not None) or isinstance(last_key, int)) and isinstance(parent[last_key], dict):
-                converted = [parent[last_key].get(str(i), None) for i in range(len(parent[last_key]))]
-                parent[last_key] = converted
+                # convert from dict to list if possible
+                converted = []
+                keys = parent[last_key].keys()
+                for i in range(len(keys)):
+                    if str(i) not in keys or parent[last_key][str(i)] is None:
+                        # Key not found
+                        break
+                    converted.append(parent[last_key][str(i)])
+                else:
+                    # overwrite the key only if the conversion was successful
+                    parent[last_key] = converted
                 
 
+        if (len(all_layers) > 1):
+            result['_layers'] = all_layers
+
         # Return the result dictionary
         return result
 
@@ -222,7 +281,7 @@ class HELML:
 
         Args:
             value: The value to be encoded.
-            spc_ch: The space character used for encoding. Default " ".
+            spc_ch: The space character used for encoding.
 
         Returns:
             str: The encoded HELML string.
@@ -230,18 +289,10 @@ class HELML:
 
         value_type = type(value).__name__
         if value_type == "str":
-            reg_str = r"^[ -~]*$"
-            if spc_ch == "_": # for url-mode
-                need_encode = "~" in value
-            else:
-                need_encode = False
-                # reg_str = r"^[\p{Print}]*$"
-
-            # if need_encode or not all(c.isprintable() for c in value) or ("_" == spc_ch and "~" in value):
-            if need_encode or not re.match(reg_str, value, flags=re.UNICODE) or ("_" == spc_ch and "~" in value):
+            if not value or not (re.match(r'^[ -}]+$', value) if spc_ch == '_' else re.match(r'^[^\x00-\x1F\x7E-\xFF]+$', value)):
                 # if the string contains special characters, encode it in base64
-                return HELML.base64url_encode(value)
-            elif not value or value[0] == spc_ch or value[-1] == spc_ch or value[-1] == ' ':
+                return '-' + HELML.base64Uencode(value)
+            elif value[0] == spc_ch or value[-1] == spc_ch or value[-1] == ' ':
                 # for empty strings or those that have spaces at the beginning or end
                 return "'" + value + "'"
             else:
@@ -260,7 +311,7 @@ class HELML:
             elif value == '-inf':
                 value = 'NIF'
             elif spc_ch == "_": # for url-mode because float contain dot-inside
-                return HELML.base64url_encode(value)
+                return HELML.base64Uencode(value)
 
         return spc_ch * 2 + str(value)
 
@@ -285,18 +336,18 @@ class HELML:
                 # if the string starts with only one space, return the string after it
                 return encoded_value[1:]
             # if the string starts with two spaces, then it encodes a non-string value
-            encoded_value = encoded_value[2:]  # strip left spaces
-            if encoded_value in HELML.SPEC_TYPE_VALUES:
-                return HELML.SPEC_TYPE_VALUES[encoded_value]
+            sliced_value = encoded_value[2:]  # strip left spaces
+            if sliced_value in HELML.SPEC_TYPE_VALUES:
+                return HELML.SPEC_TYPE_VALUES[sliced_value]
 
-            if HELML.is_numeric(encoded_value):
+            if HELML.is_numeric(sliced_value):
                 # it's probably a numeric value
-                if '.' in encoded_value:
+                if '.' in sliced_value:
                     # if there's a decimal point, it's a floating point number
-                    return float(encoded_value)
+                    return float(sliced_value)
                 else:
                     # if there's no decimal point, it's an integer
-                    return int(encoded_value)
+                    return int(sliced_value)
 
             if HELML.CUSTOM_FORMAT_DECODER is not None:
                 encoded_value = HELML.CUSTOM_FORMAT_DECODER(encoded_value, spc_ch)
@@ -310,22 +361,36 @@ class HELML:
                 return encoded_value.encode('utf-8').decode('unicode_escape')
             except ValueError:
                 return False
+        elif first_char == '-':
+            return HELML.base64Udecode(encoded_value[1:])
 
-        # if there are no spaces or quotes at the beginning, the value should be in base64
-        try:
-            return HELML.base64url_decode(encoded_value)
-        except ValueError:
-            return encoded_value
+        # if there are no spaces or quotes or "-" at the beginning
+        if HELML.CUSTOM_FORMAT_DECODER is not None:
+            # use custom-decoder if defined
+            return HELML.CUSTOM_FORMAT_DECODER(encoded_value, spc_ch)
+
+        # fallback: will return decoded string or None
+        return HELML.base64Udecode(encoded_value)
 
     @staticmethod
-    def base64url_encode(string: str) -> str:
+    def base64Uencode(string: str) -> str:
          enc = base64.b64encode(string.encode())
          return enc.decode().rstrip("=").translate(str.maketrans('+/', '-_'))
 
     @staticmethod
-    def base64url_decode(s: str) -> str:
-        s += "=" * (4 - len(s) % 4)
-        return base64.b64decode(s.translate(str.maketrans('-_', '+/')).encode()).decode()
+    def base64Udecode(s: str) -> str:
+        l = len(s)
+        if not l:
+            return ""
+        l %= 4
+        if l:
+            s += "=" * (4 - l)
+        try:
+            decoded = base64.b64decode(s.translate(str.maketrans('-_', '+/')).encode()).decode()
+            return decoded
+        except Exception:
+            return None
+
 
     @staticmethod
     def is_numeric(value: str) -> bool:
@@ -333,4 +398,26 @@ class HELML:
             float(value)
             return True
         except ValueError:
-            return False
\ No newline at end of file
+            return False
+
+    @staticmethod
+    def num_keys_cnt(arr: Union[dict, list, tuple, set]) -> int:
+        el_count = len(arr)
+        if not el_count:
+            return 0
+        if not isinstance(arr, dict):
+            return el_count
+        
+        # check for non-empty dict only, get keys list
+        keys_list = arr.keys()
+        exp_keys_range = range(0, el_count)
+        for i in keys_list:
+            if not isinstance(i, (int, str)):
+                return 0
+            try:
+                if int(i) not in exp_keys_range:
+                    return 0
+            except ValueError:
+                return 0
+
+        return el_count
\ No newline at end of file
